/*
  This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <iostream>
#include <stdio.h>
#include <iomanip>
#include <float.h>
#include <math.h>
#include "alglibmisc.h"
#include "alglibinternal.h"
#include "linalg.h"
#include "statistics.h"
#include "dataanalysis.h"
#include "specialfunctions.h"
#include "solvers.h"
#include "optimization.h"
#include "diffequations.h"
#include "fasttransforms.h"
#include "integration.h"
#include "interpolation.h"

using namespace std;


typedef struct vpt{
    unsigned int id;
    double *signature;
    unsigned int totSignature;
    
    
    double medianDistance;
    struct vpt *filhos[2];
    
    unsigned int *ids;
    unsigned int totIds;
    double **signatures;
    
    
    //LDA
    unsigned int dimension;
    unsigned int totClasses;
    unsigned int *totPerClass;
    double *classes;
    double **outLDA;
    double **projectionCentroids;
    double error;
    
    
}VPT;



//Prototypes
unsigned int countNodesVPT(VPT *tree);
unsigned int countNodesDiscriminantsVPT(VPT *tree);
unsigned int heightTree(VPT *tree);
VPT* createVPTNode();
void showVPT(VPT *tree, FILE *f);
VPT* insertVPT(double **elts, int totElts, unsigned int cols, unsigned int *posElts, unsigned int totPosElts, double LIM_NOD);
VPT* indexVPT(double **elts, unsigned int totElts, unsigned int cols, double LIM_NOD);
void showDataset(double **v, unsigned int rows, unsigned int cols, FILE *stream);
double** readDataset(char *path, unsigned int *rows, unsigned int *cols);
unsigned int* choosePivots(int quant, double **elts, unsigned int totElts, unsigned int cols, unsigned int *posElts, unsigned int totPosElts);
double distance_double(double *s1, double *s2, unsigned int tot);
void retrieve_VPT(VPT *tree, double **output, double **outputClasses, unsigned int *totOutput, double *query, double radius, unsigned int *comps, unsigned int *votes, double**train);
void createDiscriminants(VPT *tree, double **elts, unsigned int cols);
void getDecisionBoundary2D(VPT *tree, double delta, double radiusQuery);
double mean(double *v, unsigned int t);
double standDev(double *v, unsigned int t);





//Implementations


template <typename T>
void swap(T* a, T* b) {
    T tmp;
    tmp = *a;
    *a = *b;
    *b = tmp;
}


template <typename T>
int partition(T vec[], int left, int right, unsigned int indexes[]) {
    int i, j;
    
    i = left;
    for (j = left + 1; j <= right; ++j) {
        if (vec[j] < vec[left]) {
            ++i;
            //			cout << "swap " << i << "," << j << " left " << left << " right " << right << endl;
            swap(&vec[i], &vec[j]);
            if (indexes!=NULL)
                swap(&indexes[i], &indexes[j]);
        }
    }
    swap(&vec[left], &vec[i]);
    if (indexes!=NULL)
        swap(&indexes[left], &indexes[i]);
    
    return i;
}

template <typename T>
void quickSort(T vec[], int left, int right, unsigned int indexes[]) {
    //	cout << "quicksort " <<  " left " << left << " right " << right << endl;
    int r;
    if (right > left) {
        r = partition(vec, left, right, indexes);
        quickSort(vec, left, r - 1, indexes);
        quickSort(vec, r + 1, right, indexes);
    }
}




double distance_double(double *s1, double *s2, unsigned int tot){
    double dist=0, x;
    for (unsigned int i=0;i<tot; i++){
        //Euclidean
        x=s1[i]-s2[i];
        dist+=x*x;
    }
    return(sqrt(dist));
}

double** readDataset(char *path, unsigned int *rows, unsigned int *cols){
    double **dataset=NULL;
    *rows=0;
    *cols=0;
    
    FILE *f=fopen(path,"r");
    if (f==NULL){
        std::printf("%s not found!\n",path);
        
        exit(1);
    }
    
    fread(rows, sizeof(unsigned int),1,f);
    fread(cols, sizeof(unsigned int),1,f);
    dataset=(double**)malloc((*rows)*sizeof(double*));
    for (unsigned int i=0; i<*rows; i++)
        dataset[i]=(double*)malloc((*cols)*sizeof(double));
    for (int i=0;i<*rows;i++){
        for (int j=0;j<*cols;j++){
            fread(&dataset[i][j], sizeof(double),1,f);
        }
    }
    
    
    fclose(f);
    return(dataset);
}


void showDataset(double **v, unsigned int rows, unsigned int cols, FILE *stream){
    for (int i=0;i<rows;i++){
        for (int j=0;j<cols;j++)
            fprintf(stream,"%.2f ",v[i][j]);
        fprintf(stream,"\n");
    }
}



void createDiscriminants(VPT *tree, double **elts, unsigned int cols){
    //		cout<<"createDiscriminant() "<<tree->totIds<<"\n";
    if (tree==NULL)
        return;
    
    
    if (tree->totIds>0){
        
        /*
         cout<<"Vai criar Discriminant para "<<tree->totIds<<"\n";
         for (int i=0; i<tree->totIds; i++)
         cout<<tree->ids[i]<<" ";
         cout<<"\n";
         */
        
        tree->dimension=cols-1;
        
        for (unsigned int i=0; i<tree->totIds; i++){
            bool esta=false;
            for (unsigned int k=0; k<tree->totClasses; k++){
                if (tree->classes[k]==elts[tree->ids[i]][tree->dimension]){
                    tree->totPerClass[k]++;
                    esta=true;
                    break;
                }
            }
            if (!esta){
                tree->classes=(double*)realloc(tree->classes,++tree->totClasses*sizeof(double));
                tree->totPerClass=(unsigned int*)realloc(tree->totPerClass,tree->totClasses*sizeof(unsigned int));
                tree->classes[tree->totClasses-1]=elts[tree->ids[i]][tree->dimension];
                tree->totPerClass[tree->totClasses-1]=1;
            }
        }
        
        
        
        if (tree->totClasses>=2){
            
            
            
            double **centroids=(double**)malloc(tree->totClasses*sizeof(double*));
            for (unsigned int k=0; k<tree->totClasses; k++)
                centroids[k]=(double*)calloc(tree->dimension, sizeof(double));
            
            
            for (unsigned int i=0; i<tree->totIds; i++){
                for (unsigned int k=0; k<tree->totClasses; k++){
                    if (elts[tree->ids[i]][tree->dimension]==tree->classes[k]){
                        for (unsigned int d=0;d<tree->dimension; d++)
                            centroids[k][d]+=elts[tree->ids[i]][d];
                        break;
                    }
                }
            }
            for (unsigned int k=0; k<tree->totClasses; k++)
                for (unsigned int d=0;d<tree->dimension; d++)
                    centroids[k][d]/=tree->totPerClass[k];
            
            
            /*
             cout<<"Tot Classes: "<<tree->totClasses<<" Tot Per Class:\n";
             for(unsigned int i=0;i<tree->totClasses;i++)
             cout<<"CLA: "<<tree->classes[i]<<" Tot: "<<tree->totPerClass[i]<<" ";
             cout << std::endl;
             */
            
            
            
            alglib::real_2d_array data;
            data.setlength(tree->totIds,tree->dimension+1);
            alglib::ae_int_t info;
            alglib::real_2d_array output;
            alglib::ae_int_t nPoints=tree->totIds;
            alglib::ae_int_t nVars=tree->dimension;
            alglib::ae_int_t nClasses=tree->totClasses;
            
            for(unsigned int i=0;i<tree->totIds;i++) {
                for(unsigned int j=0;j<tree->dimension;j++)
                    data[i][j]=elts[tree->ids[i]][j];
                
                for (unsigned int k=0; k<tree->totClasses; k++)
                    if (tree->classes[k]==elts[tree->ids[i]][tree->dimension]){
                        data[i][tree->dimension]=k;
                    }
            }
            
            
            
            
            
            
            //			cout<<"fisherdlan("<<tree->totIds<<","<<tree->dimension<<","<<tree->totClasses<<")\n";
            alglib::fisherldan(data,nPoints,nVars,nClasses,info,output);
            
            /*
             std::cout <<"(Out LDAN) INFO: "<<info<<"\n WEIGHTS: ";
             for (unsigned int i=0; i<tree->totClasses-1; i++){
             std::cout<<"Vector "<<i<<"\n";
             for(unsigned int j=0;j<tree->dimension;j++)
             std::cout<<output[i][j]<<" ";
             std::cout<<"\n";
             }
             */
            
            
            
            tree->outLDA=(double**)malloc((tree->totClasses-1)*sizeof(double*));
            for (unsigned int k=0; k<tree->totClasses-1; k++)
                tree->outLDA[k]=(double*)malloc(tree->dimension*sizeof(double));
            
            
            
            
            for(unsigned int k=0;k<tree->totClasses-1;k++)
                for (unsigned int i=0; i<tree->dimension; i++)
                    tree->outLDA[k][i]=output[i][k];
            
            
            
            tree->projectionCentroids=(double**)malloc(tree->totClasses*sizeof(double*));
            for (unsigned int k=0; k<tree->totClasses; k++)
                tree->projectionCentroids[k]=(double*)calloc((tree->totClasses-1),sizeof(double));
            
            
            for (unsigned int i=0; i<tree->totClasses; i++){
                for (unsigned int k=0; k<tree->totClasses-1; k++){
                    for (unsigned int j=0; j<tree->dimension; j++){
                        tree->projectionCentroids[i][k]+=centroids[i][j]*tree->outLDA[k][j];
                    }
                }
            }
            
            for (unsigned int k=0; k<tree->totClasses; k++)
                free(centroids[k]);
            free(centroids);
            
            double *projected=(double*)calloc(tree->totClasses-1,sizeof(double));
            for (unsigned int i=0; i<tree->totIds; i++){
                
                for (unsigned int k=0; k<tree->totClasses-1; k++){
                    projected[k]=0;
                    for (unsigned int d=0; d<tree->dimension; d++){
                        projected[k]=projected[k]+elts[tree->ids[i]][d]*tree->outLDA[k][d];
                    }
                }
                
                double minDist=DBL_MAX, predictedClass=0;
                for (unsigned int k=0; k<tree->totClasses; k++){
                    double dist=distance_double(projected, tree->projectionCentroids[k], tree->totClasses-1);
                    if (dist<minDist){
                        predictedClass=tree->classes[k];
                        minDist=dist;
                    }
                }
                if (predictedClass!=elts[tree->ids[i]][tree->dimension]){
                    tree->error++;
                }
            }
            free(projected);
            tree->error/=tree->totIds;
        }
        return;
    }
    
    
    createDiscriminants(tree->filhos[0], elts,cols);
    createDiscriminants(tree->filhos[1], elts,cols);
    
}


VPT* indexVPT(double **elts, unsigned int totElts, unsigned int cols, double LIM_NOD){
    unsigned int *posElts=(unsigned int*)malloc(totElts*sizeof(unsigned int));
    for (unsigned int i=0; i< totElts; i++)
        posElts[i]=i;
    VPT *vptTree=insertVPT(elts, totElts, cols, posElts, totElts, LIM_NOD);
    free(posElts);
    
    
    
    createDiscriminants(vptTree, elts, cols);
    
    
    
    return(vptTree);
}


VPT* createVPTNode(){
    VPT* ret=(VPT*)malloc(sizeof(VPT));
    ret->id=0;
    ret->signature=NULL;
    ret->totSignature=0;
    
    ret->medianDistance=0;
    
    
    ret->filhos[0]=NULL;
    ret->filhos[1]=NULL;
    
    
    ret->ids=NULL;
    ret->totIds=0;
    ret->signatures=NULL;
    
    ret->dimension=0;
    ret->totClasses=0;
    ret->totPerClass=NULL;
    ret->classes=NULL;
    ret->outLDA=NULL;
    ret->projectionCentroids=NULL;
    ret->error=0;
    
    
    
    return(ret);
}





unsigned int* choosePivots(int quant, double **elts, unsigned int totElts, unsigned int cols, unsigned int *posElts, unsigned int totPosElts){
    
    if (totPosElts<quant){
        return(NULL);
    }
    srand((unsigned int)time(NULL));
    unsigned int *ret=(unsigned int*)malloc(quant*sizeof(unsigned int));
    
    
    
    
    
    
    //////////////////////////////////////////////////////////////////////////////
    //Criterium: maximize the minimum distances between a candidate and all remaining others
    unsigned int totPossibilities=0, *possibilities=NULL;
    double *distances=NULL;
    totPossibilities=totPosElts>15?15:totPosElts;
    possibilities=(unsigned int*)malloc(totPossibilities*sizeof(unsigned int));
    distances=(double*)calloc(totPossibilities,sizeof(double));
    
    for (unsigned int i=0; i<totPossibilities; i++)
        distances[i]=UINT_MAX;
    
    
    for (unsigned int i=0; i<totPossibilities; i++){
        possibilities[i]=posElts[rand()%totPosElts];
        unsigned int j;
        for (j=0; j<i; j++)
            if (possibilities[i]==possibilities[j])
                break;
        if (j<i)
            i--;
    }
    
    
    for (unsigned int i=0; i<totPossibilities; i++){
        for (unsigned int j=0; j<totPossibilities; j++){
            if (possibilities[i]!=possibilities[j]){
                unsigned int d=distance_double(elts[possibilities[i]], elts[possibilities[j]], cols-1);
                distances[i]=d<distances[i]?d:distances[i];
            }
        }
    }
    quickSort(distances, 0, (int) totPossibilities-1, possibilities);
    
    for (unsigned int i=0; i<quant; i++){
        ret[i]=possibilities[totPossibilities-1-i];
    }
    free(distances);
    free(possibilities);
    //////////////////////////////////////////////////////////////////////////////
    
    
    
    /*
     
     //////////////////////////////////////////////////////////////////////////////
     //Criterium: minimize entropy of class separation
     unsigned int totPossibilities=0, *possibilities=NULL, *possibilitiesAux=NULL;
     double *distances=NULL, *entropies=NULL;
     totPossibilities=totPosElts>15?15:totPosElts;
     possibilities=(unsigned int*)malloc(totPossibilities*sizeof(unsigned int));
     distances=(double*)calloc(totPossibilities,sizeof(double));
     entropies=(double*)calloc(totPossibilities,sizeof(double));
     possibilitiesAux=(unsigned int*)malloc((totPossibilities-1)*sizeof(unsigned int));
     
     
     
     for (unsigned int i=0; i<totPossibilities; i++){
     possibilities[i]=posElts[rand()%totPosElts];
     unsigned int j;
     for (j=0; j<i; j++)
     if (possibilities[i]==possibilities[j])
     break;
     if (j<i)
     i--;
     }
     
     
     
     for (unsigned int i=0; i<totPossibilities; i++){
     unsigned int index=0;
     for (unsigned int j=0; j<totPossibilities; j++){
     possibilitiesAux[index++]=possibilities[j];
     unsigned int d=distance_double(elts[possibilities[i]], elts[possibilities[j]], cols-1);
     distances[j]=d;
     }
     quickSort(distances, 0, (int) totPossibilities-1, possibilitiesAux);
     
     
     entropies[i]=0;
     double *classesL=NULL, *classesR=NULL;
     unsigned int totClassesL=0, totClassesR=0;
     unsigned int *totPerClassL=NULL, *totPerClassR=NULL;
     
     
     for (unsigned int j=0; j<=(unsigned int)(((double)totPossibilities-1e-10)/2); j++){
     bool esta=false;
     for (unsigned int k=0; k<totClassesL; k++){
     if (elts[possibilitiesAux[j]][cols-1]==classesL[k]){
     totPerClassL[k]++;
     esta=true;
     break;
     }
     }
     if (!esta){
     classesL=(double*)realloc(classesL,++totClassesL*sizeof(double));
     totPerClassL=(unsigned int*)realloc(totPerClassL,totClassesL*sizeof(double));
     classesL[totClassesL-1]=elts[possibilitiesAux[j]][cols-1];
     totPerClassL[totClassesL-1]=1;
     }
     }
     for (unsigned int j=(unsigned int)(((double)totPossibilities-1e-10)/2)+1; j<totPossibilities; j++){
     bool esta=false;
     for (unsigned int k=0; k<totClassesR; k++){
     if (elts[possibilitiesAux[j]][cols-1]==classesR[k]){
     totPerClassR[k]++;
     esta=true;
     break;
     }
     }
     if (!esta){
     classesR=(double*)realloc(classesR,++totClassesR*sizeof(double));
     totPerClassR=(unsigned int*)realloc(totPerClassR,totClassesR*sizeof(double));
     classesR[totClassesR-1]=elts[possibilitiesAux[j]][cols-1];
     totPerClassR[totClassesR-1]=1;
     }
     }
     
     unsigned int totL=(unsigned int)(((double)totPossibilities-1e-10)/2)+1;
     unsigned int totR=totPossibilities-(unsigned int)(((double)totPossibilities-1e-10)/2)+1;
     
     for (unsigned int k=0; k<totClassesL; k++)
     entropies[i]+=(double)totPerClassL[k]/totL*log2((double)totPerClassL[k]/totL);
     for (unsigned int k=0; k<totClassesR; k++)
     entropies[i]+=(double)totPerClassR[k]/totR*log2((double)totPerClassR[k]/totR);
     
     
     free(classesL);
     free(classesR);
     free(totPerClassL);
     free(totPerClassR);
     }
     quickSort(entropies, 0, (int) totPossibilities, possibilities);
     
     for (unsigned int i=0; i<quant; i++){
     ret[i]=possibilities[totPossibilities-1-i];
     //		cout<<" Select "<<possibilities[totPossibilities-1-i]<<"\n";
     }
     
     
     free(distances);
     free(possibilities);
     free(entropies);
     free(possibilitiesAux);
     //////////////////////////////////////////////////////////////////////////////
     */
    
    
    
    
    return(ret);
    
}


VPT* insertVPT(double **elts, int totElts, unsigned int cols, unsigned int *posElts, unsigned int totPosElts, double LIM_NOD){
    
    
    
    if (totPosElts==0)
        return(NULL);
    
    VPT *tree=createVPTNode();
    
    //	if ((((double)totPosElts)/totElts<=LIM_NOD)||(totPosElts<=1)){
    if (totPosElts<=LIM_NOD){
        
        
        tree->totIds=totPosElts;
        tree->ids=(unsigned int*)realloc(tree->ids,tree->totIds*sizeof(unsigned int));
        tree->totSignature=cols;
        tree->signatures=(double**)realloc(tree->signatures,tree->totIds*sizeof(double*));
        for (unsigned int i=0; i<tree->totIds; i++){
            tree->ids[i]=posElts[i];
            tree->signatures[i]=(double*)malloc(cols*sizeof(double));
            for (unsigned int j=0; j<cols; j++)
                tree->signatures[i][j]=elts[posElts[i]][j];
        }
        return(tree);
    }
    
    
    
    
    
    unsigned int* posPivotElts=choosePivots(1,elts, totElts, cols, posElts, totPosElts);
    double *distances=NULL;
    unsigned int *eltsDistances=NULL;
    unsigned int totDistances=0;
    
    
    
    
    tree->id=posPivotElts[0];
    tree->signature=(double*)malloc(cols*sizeof(double));
    tree->totSignature=cols;
    for (unsigned int i=0;i<tree->totSignature; i++)
        tree->signature[i]=elts[posPivotElts[0]][i];
    
    
    
    
    if (totPosElts==1){
        free(posPivotElts);
        free(distances);
        free(eltsDistances);
        tree->medianDistance=0;
        return(tree);
    }
    
    
    for (unsigned int i=0; i<totPosElts; i++){
        if (posPivotElts[0]==posElts[i])			//Whether to pass (or not) the pivot to nodes below
            continue;
        distances=(double *)realloc(distances,++totDistances*sizeof(double));
        eltsDistances=(unsigned int *)realloc(eltsDistances,totDistances*sizeof(unsigned int));
        distances[totDistances-1]=distance_double(tree->signature, elts[posElts[i]], tree->totSignature-1);
        eltsDistances[totDistances-1]=posElts[i];
    }
    quickSort(distances, 0, (int) totDistances-1, eltsDistances);
    
    
    
    double posMedian=((double)(totDistances+1))/2-1;
    tree->medianDistance=(distances[(int)floor(posMedian)]+distances[(int)ceil(posMedian)])/2;
    
    
    
    
    tree->filhos[0]=insertVPT(elts, totElts, cols, eltsDistances,floor(posMedian+1), LIM_NOD);
    tree->filhos[1]=insertVPT(elts, totElts, cols, eltsDistances+(int)ceil(posMedian),floor(totDistances-posMedian), LIM_NOD);
    
    
    
    free(posPivotElts);
    free(distances);
    free(eltsDistances);
    
    
    return(tree);
}





void showVPT(VPT *tree, FILE *f){
    if (tree==NULL)
        return;
    fprintf(f, "<ID: %u>\n",tree->id);
    fprintf(f, "<median: %f>\n",tree->medianDistance);
    fprintf(f, "<error: %f>\n",tree->error);
    fprintf(f, "	<tot ids: %u>\n",tree->totIds);
    for (unsigned int i=0; i< tree->totIds; i++)
        fprintf(f,"%u ",tree->ids[i]);
    fprintf(f,"\n");
    fprintf(f," <tot classes: %u>\n",tree->totClasses);
    for(unsigned int i=0; i<tree->totClasses; i++){
        fprintf(f,"CLASS %.0f Tot %u ",tree->classes[i],tree->totPerClass[i]);
    }
    fprintf(f,"\n");
    showVPT(tree->filhos[0], f);
    showVPT(tree->filhos[1], f);
    fprintf(f, "</ID: %u>\n",tree->id);
}


unsigned int countNodesVPT(VPT *tree){
    if (tree==NULL)
        return((unsigned int)0);
    return((unsigned int)1+countNodesVPT(tree->filhos[0])+countNodesVPT(tree->filhos[1]));
}

unsigned int countNodesDiscriminantsVPT(VPT *tree){
    if (tree==NULL)
        return((unsigned int)0);
    if (tree->totIds>0)
        return(1);
    return(countNodesDiscriminantsVPT(tree->filhos[0])+countNodesDiscriminantsVPT(tree->filhos[1]));
}


VPT* clearVPTTree(VPT *tree){
    if (tree==NULL)
        return(NULL);
    
    
    free(tree->signature);
    free(tree->ids);
    for (unsigned int i=0; i<tree->totIds; i++)
        free(tree->signatures[i]);
    free(tree->signatures);
    
    
    
    free(tree->totPerClass);
    free(tree->classes);
    
    
    for (int i=0; i<(int)tree->totClasses-1; i++)
        free(tree->outLDA[i]);
    free(tree->outLDA);
    
    if (tree->totClasses>=2)
        for (unsigned int i=0; i<tree->totClasses; i++)
            free(tree->projectionCentroids[i]);
    free(tree->projectionCentroids);
    
    
    
    
    tree->filhos[0]=clearVPTTree(tree->filhos[0]);
    tree->filhos[1]=clearVPTTree(tree->filhos[1]);
    
    
    
    
    
    return(NULL);
}








void retrieve_VPT(VPT *tree, double **output, double **outputClasses, unsigned int *totOutput, double *query, double radius, unsigned int *comps, unsigned int *votes, double**train){
    //	printf("retrieve_VPT() query=(%.2f,%.2f)\n",query[0],query[1]);
    *comps=0;
    if (tree==NULL)
        return;
    *comps=1;
    if (tree->totIds>0){
        //If it is a leave
        *votes=*votes+1;
        
        unsigned int predictedClass=0;
        
        /*
         cout<<"Voting...[";
         for (int i=0;i<tree->totIds; i++)
         printf("(%.2f, %.2f)[%.0f] ",train[tree->ids[i]][0],train[tree->ids[i]][1],train[tree->ids[i]][2]);
         printf("\n");
         
         cout<<"Tot Classes Node: "<<tree->totClasses<<" TotPer Class:";
         for (int i=0;i<tree->totClasses; i++)
         printf("[%.0f]%d ",tree->classes[i],tree->totPerClass[i]);
         printf("\n");
         */
        
        if (tree->totClasses<2){
            predictedClass=tree->classes[0];
        }else{
            
            double *projected=(double*)calloc(tree->totClasses-1,sizeof(double));
            for (unsigned int k=0; k<tree->totClasses-1; k++)
                for (unsigned int d=0; d<tree->dimension; d++)
                    projected[k]=projected[k]+query[d]*tree->outLDA[k][d];
            
            
            double minDist=DBL_MAX;
            
            for (unsigned int k=0; k<tree->totClasses; k++){
                double dist=distance_double(projected, tree->projectionCentroids[k], tree->totClasses-1);
                if (dist<minDist){
                    predictedClass=tree->classes[k];
                    minDist=dist;
                }
            }
            free(projected);
            
        }
        
        
        
        for (unsigned int i=0;i<*totOutput; i++){
            if ((*outputClasses)[i]==predictedClass){
                (*output)[i]+=(1-tree->error);
                //				cout<<"Voting class "<<predictedClass<<" votes "<<(1-tree->error)<<"\n";
                return;
            }
        }
        
        *output=(double*)realloc(*output, (++(*totOutput))*sizeof(double));
        (*output)[*totOutput-1]=(1-tree->error);
        *outputClasses=(double*)realloc(*outputClasses, (*totOutput)*sizeof(double));
        (*outputClasses)[*totOutput-1]=predictedClass;
        //		cout<<"Voting class "<<predictedClass<<" votes "<<(1-tree->error)<<"\n";
        return;
    }
    
    
    
    double dist=distance_double(tree->signature, query, tree->totSignature-1);
    //	dist/=(tree->totSignature-1);
    
    //	printf("Dist (%.2f,%.2f)<->(%.2f,%.2f)=%f\n",tree->signature[0],tree->signature[1],query[0],query[1],dist);
    
    if (dist-radius<=tree->medianDistance){
        //			std::cout<<"["<<tree->id<<"]("<<train[tree->id][0]<<","<<train[tree->id][1]<<") Goes Left: "<<dist<<"-"<<radius<<"<="<<tree->medianDistance<<"\n";
        unsigned int compsT=0;
        retrieve_VPT(tree->filhos[0], output, outputClasses, totOutput, query, radius, &compsT, votes,train);
        *comps=*comps+compsT;
    }
    
    if (dist+radius>=tree->medianDistance){
        //			std::cout<<"["<<tree->id<<"]("<<train[tree->id][0]<<","<<train[tree->id][1]<<") Goes Right: "<<dist<<"+"<<radius<<">="<<tree->medianDistance<<"\n";
        unsigned int compsT=0;
        retrieve_VPT(tree->filhos[1], output, outputClasses, totOutput, query, radius, &compsT, votes,train);
        *comps=*comps+compsT;
        
    }
    //	printf("End %d\n",tree->id);
}


void getDecisionBoundary2D(VPT *tree, double delta, double radiusQuery){
    FILE *f=fopen("DecisionBoundary.txt","w");
    double input[2];
    for (double dy=1; dy>=0; dy-=delta){
        for (double dx=0; dx<=1; dx+=delta){
            
            input[0]=dx;
            input[1]=dy;
            
            double *output=NULL, *outputClasses=NULL;
            unsigned int totOutput=0;
            unsigned int votes=0, comps=0;
            
            
            
            retrieve_VPT(tree, &output, &outputClasses, &totOutput, input, radiusQuery, &comps, &votes,NULL);
            
            //			printf("-INPUT %.2f %.2f ",input[0],input[1]);
            double predictedClass=0, maxOutput=0, sumOutput=0;
            for (unsigned int j=0; j<totOutput; j++){
                //				printf("[%.0f] %.2f ",outputClasses[j], output[j]);
                sumOutput+=output[j];
                if (output[j]>maxOutput){
                    maxOutput=output[j];
                    predictedClass=outputClasses[j];
                }
            }
            //			printf("\n");
            
            free(output);
            free(outputClasses);
            
            
            fprintf(f,"%.0f ",predictedClass);
            
        }
        fprintf(f,"\n");
    }
    fclose(f);
    
}

double mean(double *v, unsigned int t){
    double ret=0;
    for (unsigned int i=0; i<t; i++)
        ret+=v[i];
    return(ret/t);
}

double standDev(double *v, unsigned int t){
    double mu=mean(v,t), ret=0;
    for (unsigned int i=0; i<t; i++)
        ret+=(mu-v[i])*(mu-v[i]);
    return(sqrt(ret/t));
}

unsigned int heightTree(VPT *tree){
    unsigned int hl=0, hr=0;
    if (tree==NULL)
        return(0);
    hl=heightTree(tree->filhos[0]);
    hr=heightTree(tree->filhos[1]);
    return(1+max(hl,hr));
}



int main (int argc, char * const argv[]) {
    
    
    if (argc<3){
        printf("ERROR:! Usage: VPC learningFile.dat testFile.dat\n");
        exit(1);
    }
        
    
    
    //PERFORMANCE
    unsigned int LIMIT_NODES=2;
    double RADIUS_QUERY=0.15;
    
    
    
    unsigned int rowsTrain, colsTrain, rowsTest, colsTest;
    

    
    double **datasetTrain=readDataset(argv[1], &rowsTrain, &colsTrain);
    cout<<"Training Readed "<<rowsTrain<<" rows "<<colsTrain<<" cols...\n";
    
    double **datasetTest=readDataset(argv[2], &rowsTest, &colsTest);
    cout<<"Test Readed "<<rowsTest<<" rows "<<colsTest<<" cols...\n";
    
    
    
    
    VPT *tree=indexVPT(datasetTrain, rowsTrain, colsTrain,LIMIT_NODES);
    //				showVPT(tree, stdout);
    cout<<"Tot Nodes tree "<<countNodesVPT(tree)<<" Tot Discriminants "<<countNodesDiscriminantsVPT(tree)<<" Height tree "<<heightTree(tree)<<"\n";
    
    
    
    
    double errors=0;
    unsigned int comps, votes=0;
    double avgVotes=0;
    
    
    
    for (unsigned int i=0;i<rowsTest; i++){
        
        
        double *output=NULL, *outputClasses=NULL;
        unsigned int totOutput=0;
        votes=0;
        
        
        
        retrieve_VPT(tree, &output, &outputClasses, &totOutput, datasetTest[i], RADIUS_QUERY, &comps, &votes, datasetTrain);
        
        
        double predictedClass=0, maxOutput=0, sumOutput=0;
        for (unsigned int j=0; j<totOutput; j++){
            sumOutput+=output[j];
            if (output[j]>maxOutput){
                maxOutput=output[j];
                predictedClass=outputClasses[j];
            }
        }
        
        
        
        free(output);
        free(outputClasses);
        
        
        
        
        
        avgVotes+=votes;
        
        if (datasetTest[i][colsTest-1]!=predictedClass)
            errors++;
        
    }
    
    avgVotes/=rowsTest;
    errors=errors/rowsTest;
    
    cout<<"++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
    cout<<"Error rate "<<errors<<", Avg. Votes "<<avgVotes<<"\n";
    cout<<"++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
    
    
    
    tree=clearVPTTree(tree);
    
    
    
    //Free training and test data sets
    for (unsigned int i=0;i<rowsTrain; i++)
        free(datasetTrain[i]);
    free(datasetTrain);
    
    for (unsigned int i=0;i<rowsTest; i++)
        free(datasetTest[i]);
    free(datasetTest);
    
    exit(0);
}





